package cn.jlsxwkj.moudles.chat;

import cn.dev33.satoken.SaManager;
import cn.hutool.core.lang.Tuple;
import cn.hutool.crypto.digest.MD5;
import cn.jlsxwkj.common.config.UserConfig;
import cn.jlsxwkj.common.exception.CustomException;
import cn.jlsxwkj.common.exception.FileTypeDoesNotSupportException;
import cn.jlsxwkj.common.exception.InsertFailException;
import cn.jlsxwkj.common.exception.UnknownException;
import cn.jlsxwkj.common.reader.ParagraphDocReader;
import cn.jlsxwkj.common.reader.ParagraphOcrReader;
import cn.jlsxwkj.common.reader.ParagraphTextReader;
import cn.jlsxwkj.common.utils.Log;
import cn.jlsxwkj.common.utils.MergeDocuments;
import cn.jlsxwkj.moudles.chathistory.ChatHistoryService;
import jakarta.annotation.Resource;
import lombok.Data;
import org.reactivestreams.Subscription;
import org.springframework.ai.chat.messages.*;
import org.springframework.ai.chat.model.ChatResponse;
import org.springframework.ai.chat.prompt.Prompt;
import org.springframework.ai.document.Document;
import org.springframework.ai.document.DocumentReader;
import org.springframework.ai.embedding.EmbeddingModel;
import org.springframework.ai.ollama.OllamaChatModel;
import org.springframework.ai.vectorstore.SearchRequest;
import org.springframework.ai.vectorstore.VectorStore;
import org.springframework.stereotype.Service;
import org.springframework.web.multipart.MultipartFile;
import reactor.core.publisher.Flux;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.Objects;
import java.util.stream.Collectors;


/**
 * @author zh
 */
@Service
@Data
public class ChatService {

    private static final MD5 MD5 = cn.hutool.crypto.digest.MD5.create();
    private static final String PATH = System.getProperty("user.dir") + "/ai_doc/";
    private static final int MAX_LIST_MESSAGE_LENGTH = 3;
    private static final int MAX_MESSAGE_LENGTH = 512;
    private static List<Message> listMessage = new ArrayList<>();

    @Resource
    private VectorStore vectorStore;
    @Resource
    private EmbeddingModel embeddingModel;
    @Resource
    private OllamaChatModel ollamaChatModel;
    @Resource
    private ChatHistoryService chatHistoryService;
    @Resource
    private UserConfig userConfig;

    private StringBuffer chatMessage;
    private Subscription subscription;

    /**
     * 使用spring ai解析txt文档并保存至 pg
     *
     * @param file 保存的文件
     * @return 保存状态
     */
    public String uploadDocument(MultipartFile file) throws CustomException {
        Tuple supportFile = isSupportFile(file);
        File saveFile = supportFile.get(0);
        File fileName = supportFile.get(1);
        DocumentReader reader = supportFile.get(3);
        if (!saveFile.exists()) {
            try {
                file.transferTo(saveFile);
                vectorStore.add(reader.get());
                return "上传完成";
            } catch (Exception e) {
                boolean delete = saveFile.delete();
                if (!delete) {
                    Log.warn(this.getClass(), "删除文件失败: {}", fileName);
                    throw new UnknownException();
                }
            }
        }
        return "上传完成";
    }

    /**
     * 根据关键词搜索向量库
     *
     * @param keyword 关键词
     * @return 文本内容
     */
    public String search(String keyword) {
        return mergeDocuments(vectorStore.similaritySearch(SearchRequest.query(keyword).withSimilarityThreshold(0.5)));
    }

    /**
     * 问答流,根据输入内容回答
     *
     * @param message 输入内容
     * @return 回答内容
     */
    public Flux<String> chatStream(String message) {
        String context = search(message);
        return stream(message, new SystemMessage(context));
    }

    /**
     * 问答总结流, 根据给定的文档总结内容
     *
     * @param message 输入内容
     * @param file    文件
     * @return 内容总结
     */
    public Flux<String> chatSummaryStream(String message, MultipartFile file) throws CustomException {
        embeddingModel.embed("");
        Tuple supportFile = isSupportFile(file);
        String fileName = supportFile.get(1);
        String fileTypeCn = supportFile.get(2);
        DocumentReader reader = supportFile.get(3);
        String context = fileTypeCn + " :: " + fileName + " ====> " + mergeDocuments(reader.get()) + "\n" + "↑ " + message;
        UserMessage userMessage = new UserMessage(context);
        Log.info(this.getClass(), context);
        return stream(message, userMessage);
    }

    /**
     * 取消订阅
     */
    public void stopChat() {
        subscription.cancel();
    }

    private String mergeDocuments(List<Document> documents) {
        return MergeDocuments.mergeDocuments(documents).stream()
                .map(Document::getContent).collect(Collectors.joining("\n"));
    }

    /**
     * 获取prompt
     *
     * @param message 提问内容
     * @param context 提示词
     * @return 提示词
     */
    private Prompt getChatPrompt2String(String message, Message context) {
        if (listMessage.size() > MAX_LIST_MESSAGE_LENGTH) {
            listMessage = listMessage.subList(listMessage.size() - MAX_LIST_MESSAGE_LENGTH, listMessage.size())
                    .stream()
                    .map(msg -> {
                        if (msg.getMessageType().equals(MessageType.ASSISTANT)) {
                            String content = msg.getContent();
                            if (content.length() > MAX_MESSAGE_LENGTH) {
                                return new AssistantMessage(content.substring(0, MAX_MESSAGE_LENGTH));
                            }
                        }
                        return msg;
                    }).collect(Collectors.toList());
        }
        listMessage.add(new SystemMessage(userConfig.getSysMessage()));
        if (context.getMessageType().equals(MessageType.SYSTEM)) {
            listMessage.add(new UserMessage(message));
            listMessage.add(context);
        }
        if (context.getMessageType().equals(MessageType.USER)) {
            return new Prompt(context);
        }
        return new Prompt(listMessage);
    }

    /**
     * chat stream 封装
     *
     * @param message 提问内容
     * @param context 提示词
     * @return flux
     */
    private Flux<String> stream(String message, Message context) {
        String userId = SaManager.getStpLogic("").getLoginIdAsString();
        chatMessage = new StringBuffer();
        Flux<ChatResponse> stream = ollamaChatModel.stream(getChatPrompt2String(message, context));
        return stream.doOnSubscribe(subscription -> this.subscription = subscription).map(
                response -> {
                    String str = response.getResult().getOutput().getContent();
                    chatMessage.append(str);
                    return str.replace(" ", "  ")
                            .replace("\t", "    ");
                }).concatWithValues("<{完成}>")
                .onErrorComplete()
                .doOnComplete(() -> {
                    String chatMsg = chatMessage.toString();
                    listMessage.add(new AssistantMessage(chatMsg));
                    try {
                        chatHistoryService.insert(userId, message, chatMsg);
                    } catch (InsertFailException e) {
                        Log.warn(this.getClass(), e.getMessage());
                    }
                    Log.info(this.getClass(), "{} user message ==> {}", userId, message);
                    Log.info(this.getClass(), "{} chat message ==> {}", userId, chatMsg);
                });
    }

    /**
     * 封装文件
     *
     * @param file 文件流
     * @return 元组
     */
    private Tuple isSupportFile(MultipartFile file) throws CustomException {
        byte[] bytes = new byte[0];
        try {
            bytes = file.getBytes();
        } catch (IOException e) {
            e.printStackTrace();
        }
        String[] split = Objects.requireNonNull(file.getOriginalFilename()).split("\\.");
        String fileType = split[split.length - 1].toLowerCase(Locale.ROOT);
        File savePath = new File(PATH);
        File saveFile = new File(PATH + ChatService.MD5.digestHex(bytes) + "." + fileType);
        String fileName = saveFile.getName();
        String fileTypeCn;
        if (!savePath.exists()) {
            if (!savePath.mkdirs()) {
                Log.warn(this.getClass(), "创建文件夹失败: " + PATH);
            }
        }
        DocumentReader reader;
        switch (fileType) {
            case "txt" -> {
                reader = new ParagraphTextReader(bytes, fileName, 5);
                fileTypeCn = "文本";
            }
            case "doc", "docx" -> {
                reader = new ParagraphDocReader(bytes, fileName, 5);
                fileTypeCn = "文档";
            }
            case "jpg", "png" -> {
                reader = new ParagraphOcrReader(bytes, fileName, userConfig.getCnocrUrl(), 5);
                fileTypeCn = "图片";
            }
            default -> throw new FileTypeDoesNotSupportException("暂不支持的文件类型: " + fileType);
        }
        return new Tuple(saveFile, fileName, fileTypeCn, reader);
    }
}
